/*****************************************************************************
 *
 *                       Cinepak for SATURN Player by SOJ
 *                         usage sample program for SGL
 *
 * Using the file system to play a movie file on a CD
 * Sample of playing a movie file on a CD in 16 million colors on VDP2 using the file system
 * Movie file name is "SAMPLE.CPK"
 *
 *                      Copyright(c) 1994,1995 SEGA
 *
 *****************************************************************************/

//#define USE_STM // Define this when using the Stream System : STM not working at the moment

/*--------------------------------------------------------------------------*/
/*                 Include                  */
/*--------------------------------------------------------------------------*/
#include "sgl.h"
#include "sgl_cd.h"
#include "sega_cpk.h"
#include "sega_snd.h"
#include "sega_int.h"
#include "sega_tim.h"
#ifdef USE_STM
    #include "sega_stm.h"
#endif

/*--------------------------------------------------------------------------*/
/*                 Define                   */
/*--------------------------------------------------------------------------*/
#define WIDTH_H 320
#define WIDTH_V 224

// VRAM cycle pattern (bank A0) register
#define CYCLE_A_REG 0x25f80010

// VRAM cycle pattern (bank B0) register
#define CYCLE_B_REG 0x25f80018

// Turn on CPU read/write mode
#define CYCLE_CPU_WRITE(reg_addr) { *((Uint32 *)(reg_addr)) = 0xeeeeeeee; }

// Set to character pattern data read mode
#define CYCLE_VDP_READ(reg_addr) { *((Uint32 *)(reg_addr)) = 0x44444444; }

#define CYCPAT_BM_READ 0x44444444
#define CYCPAT_CPU_RW 0xffffffff

#define RING_BUF_SIZ (1024L*200)

// PCM RAM transfer address and number of samples
#define PCM_ADDR ( (void*)0x25a20000 )

#define PCM_SIZE (4096L*16) // Standard PCM Size

#define START_TRG_SIZE (100*2048)

/*--------------------------------------------------------------------------*/
/*                 Extern                   */
/*--------------------------------------------------------------------------*/
extern Uint8 PauseFlag;
extern Uint32 sddrvsize;
extern Uint32 bootsndsize;
extern char sddrvstsk[];
extern char bootsnd[];

/*--------------------------------------------------------------------------*/
/*                Structure                 */
/*--------------------------------------------------------------------------*/
volatile Sint32 Switch_VBL_IN = TRUE;
volatile Sint32 Switch_timer0 = TRUE;
Uint32 movie_x;
Uint32 movie_x_size;
Uint32 movie_y_half;
Uint32 movie_y;
Uint8 *src;
Uint8 *dst;
void *cpk_vdp2_dst;

Uint32 work_ram_buf[320L * 224L * 4 / 4];

Uint32 g_movie_work[CPK_24WORK_DSIZE]; /* Work Buffer */

Uint32 g_movie_buf[RING_BUF_SIZ / sizeof(Uint32)]; /* Ring Buffer */

char *g_filenaem = "SAMPLE.CPK";
     
/*--------------------------------------------------------------------------*/
/*                GFS Routin                  */
/*--------------------------------------------------------------------------*/

// Maximum number of files in the root directory
#define MAX_DIR 40

// Maximum number of files to open simultaneously
#define OPEN_MAX 5

// Information containing the file name
GfsDirName dir_name[MAX_DIR];

// GFS Workspace
Uint8 gfs_work[GFS_WORK_SIZE(OPEN_MAX)];

// Directory Information Management Domain
GfsDirTbl dir_tbl;

#ifdef USE_STM
    Uint8 stm_work[ STM_WORK_SIZE(12, 24) ];

    // 1/2
    #define SMP_DIV2(a)   ((a) >> 1)

    // Maximum image size that can be transferred in batches during V-blank
    #define SMPCPK_VBL_COPY_MAX (288 * 144)

    void errStmFunc(void *obj, Sint32 ec)
    {
        // Error Handling
    }
#endif

// Volume Level
Sint8 g_level = 7; // Volume Audio, entre 7 (volume à fond) et 0 (silence)

// Panning Value : Center = 16 | Left = 0 | Right = 31
Sint8 g_pan = 16;

// Pan value: value specified as an argument
Sint32 g_pan_tbl[31 + 1] = {
    0x1f, 0x1e, 0x1d, 0x1c, 0x1b, 0x1a, 0x19, 0x18, 
    0x17, 0x16, 0x15, 0x14, 0x13, 0x12, 0x11, 0x10, 
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 
    0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f
};

void errGfsFunc(void *obj, Sint32 ec)
{
    // Error Handling
    slPrint("Error Gfs Func", slLocate(9, 3) );
    slPrintHex( ec, slLocate(26, 3) );
    while (-1)
    {
        slSynch();
    }
}

void fileInit(void)
{
    // Initialize the GFS file system
    GFS_DIRTBL_TYPE(&dir_tbl) = GFS_DIR_NAME;
    GFS_DIRTBL_DIRNAME(&dir_tbl) = dir_name;
    GFS_DIRTBL_NDIR(&dir_tbl) = MAX_DIR;
    Sint32 file_num = GFS_Init(OPEN_MAX, gfs_work, &dir_tbl);
    if (file_num < 0)
    {
        slPrint("Error GFS Init", slLocate(26, 3) );
        while (-1)
        {
            slSynch();
        }
    }

    GFS_SetErrFunc(errGfsFunc, NULL); // Setting the Error Function

    #ifdef USE_STM
        STM_Init(12, 24, stm_work); // Initializing the stream system
        STM_SetErrFunc(errStmFunc, NULL); // Setting the Error Function
    #endif
}

#ifdef USE_STM
    StmGrpHn grp_hd; // ID of the stream group

    StmHn stmOpen(char *fname)
    {
        StmKey key;
        // Find the file identifier from the file name
        Sint32 fid = GFS_NameToId(fname);
        STM_KEY_FN(&key) = STM_KEY_CN(&key) = STM_KEY_SMMSK(&key) = STM_KEY_SMVAL(&key) = STM_KEY_CIMSK(&key) = STM_KEY_CIVAL(&key) = STM_KEY_NONE;
        return STM_OpenFid(grp_hd, fid, &key, STM_LOOP_NOREAD);
    }

    void stmClose(StmHn fp)
    {
        STM_Close(fp);
    }
#endif

/*--------------------------------------------------------------------------*/
/*                 VBL Routin                 */
/*--------------------------------------------------------------------------*/
void smpVblIn(void)
{
    CPK_VblIn(); // Call Cinepak's VblIn routine
    Switch_VBL_IN = FALSE;
}

/*--------------------------------------------------------------------------*/
/*                 Cinepak Routin               */
/*--------------------------------------------------------------------------*/
void errCpkFunc(void *obj, Sint32 ec)
{
    slPrint("Error Cpk Func" , slLocate(9,3) );
    slPrintHex(ec, slLocate(26, 3) );
    while (-1)
    {
        slSynch();
    }
}

// Generate movie handle
CpkHn createMovie(
    #ifdef USE_STM
        StmHn stm
    #else 
        GfsHn gfs
    #endif
)
{
    CpkCreatePara para;
    CpkHeader *header;
    CpkHn cpk;

    // Generate movie handle
    CPK_PARA_WORK_ADDR(&para) = g_movie_work;
    CPK_PARA_WORK_SIZE(&para) = CPK_24WORK_BSIZE;
    CPK_PARA_BUF_ADDR(&para) = g_movie_buf;
    CPK_PARA_BUF_SIZE(&para) = RING_BUF_SIZ;
    CPK_PARA_PCM_ADDR(&para) = PCM_ADDR;
    CPK_PARA_PCM_SIZE(&para) = PCM_SIZE;
    #ifdef USE_STM
        cpk = CPK_CreateStmMovie(&para, stm);
    #else
        cpk = CPK_CreateGfsMovie(&para, gfs);
    #endif
    if (cpk == NULL)
    {
        return NULL;
    }

    // Set the transfer method to Software Transfer
    CPK_SetTrModePcm(cpk, CPK_TRMODE_CPU); 

    // Set the number of display colors to 16 millions
    CPK_SetColor(cpk, CPK_COLOR_24BIT);

    // Load the header
    // If the size of the movie is known in advance, there is no need to call CPK_Preload and CPK_GetHeader
    // In the case of a file system, it is better not to call CPK_PreloadHeader() to start playback a little earlier
    CPK_PreloadHeader(cpk);

    // Get the size of the Movie
    header = CPK_GetHeader(cpk);
    movie_x = header->width;
    movie_y = header->height;
    movie_x_size = movie_x*4;
    movie_y_half = movie_y/2;

    cpk_vdp2_dst = (Uint8*)(0x25e00000 + 4*(512*((WIDTH_V - movie_y)/2) + ((WIDTH_H - movie_x)/2)) + 4*512*(512/2 - WIDTH_V) / 2);

    // Set the Decoding Address of the movie
    CPK_SetDecodeAddr(cpk, (void *)work_ram_buf, 4 * movie_x);

    #ifdef START_TRG_SIZE
        // Set the playback start trigger size [byte]
        CPK_SetStartTrgSize(cpk, START_TRG_SIZE);
    #endif

    CPK_SetVolume(cpk, g_level);
    CPK_SetPan(cpk, g_pan_tbl[g_pan]);
    return cpk;
}

/*--------------------------------------------------------------------------*/
/*                 Sound Routin               */
/*--------------------------------------------------------------------------*/
void sndInit(void)
{
    #if 0
        slInitSound((void*)sddrvstsk, sddrvsize, (void*)bootsnd, bootsndsize);
        *(volatile Uint8 *)(0x25a004e1) = 0x00 ; /* Hand Shake Off */
    #else
        SndIniDt snd_init;

        SND_INI_PRG_ADR(snd_init) = (Uint16*)sddrvstsk;
        SND_INI_PRG_SZ(snd_init)  = (Uint16)sddrvsize;
        SND_INI_ARA_ADR(snd_init) = (Uint16*)bootsnd;
        SND_INI_ARA_SZ(snd_init)  = (Uint16)bootsndsize;
        SND_Init(&snd_init);
        SND_ChgMap(0);
    #endif
}

void timer0func(void)
{
    Switch_timer0 = FALSE;
}

void cinepak_main(void)
{
    #ifdef USE_STM
        StmHn stm;
    #endif

    volatile CpkHn cpk;
    volatile GfsHn gfs;
    volatile Uint8 restart;
    register Sint32 iu;

    slInitSystem(TV_320x224,NULL,1);
    slTVOff();
    slBitMapNbg0(COL_TYPE_1M,BM_512x256,(void *)VDP2_VRAM_A0);
    slColRAMMode(CRM16_1024);
    slScrAutoDisp(NBG0ON);
    slTVOn();
    slScrCycleSet(CYCPAT_BM_READ, CYCPAT_BM_READ, CYCPAT_BM_READ, CYCPAT_BM_READ);
    slScrPosNbg0( toFIXED(0), toFIXED((512/2 - WIDTH_V)/2) ); // x , y
    slVRAMMode((Uint16)NULL) ;
    slSynch();

    // Method for determining the correlation between multiple sets of data
    PauseFlag = -1;

    slIntFunction( (void*)smpVblIn );

    fileInit();

    sndInit();

    CPK_Init(); // Initialize Cinepak

    CPK_SetErrFunc(errCpkFunc, NULL); // Setting the Error Function

    restart = 1;

    // Timer 0 interrupt setting
    INT_SetScuFunc(INT_SCU_TIM0, timer0func);

    // Timer 0 settings
    TIM_T0_SET_CMP(WIDTH_V/2 + 2);
    TIM_T1_SET_MODE(0x101);
    TIM_T0_ENABLE();

    while (1)
    {
        if (restart)
        {
            #ifdef USE_STM
                if ((stm = stmOpen(g_filenaem)) == NULL) // Open Stream
                {
                    continue;
                }
                cpk = createMovie(stm); // Generate movie handle
            #else
                if ( (gfs = GFS_Open( GFS_NameToId(g_filenaem) ) ) == NULL) // Open File
                {
                    continue;
                }
                cpk = createMovie(gfs); // Generate movie handle
            #endif

          CPK_Start(cpk); // Movie Start
          restart = 0;
        }

        CPK_Task(cpk);

        // Check for screen display requests
        if (CPK_IsDispTime(cpk) == TRUE)
        {
            #if 1
                Switch_timer0 = TRUE;
                while(Switch_timer0);
                Switch_VBL_IN = TRUE;

                src = (Uint8*)work_ram_buf;
                dst = (Uint8*)cpk_vdp2_dst;

            #if 0
                // Handled according to specific conditions
                slScrCycleSet(CYCPAT_CPU_RW, CYCPAT_CPU_RW, CYCPAT_BM_READ, CYCPAT_BM_READ);
            #endif

            // Set bank A0 to CPU write mode
            CYCLE_CPU_WRITE(CYCLE_A_REG);

            for (iu = 0; iu < movie_y_half; iu++)
            {
                DMA_ScuMemCopy(dst, src, movie_x_size);
                src += movie_x_size;
                dst += 2048;
            }

            #if 0
                // Processed in accordance with specific file format
                slScrCycleSet(CYCPAT_BM_READ, CYCPAT_BM_READ, CYCPAT_BM_READ, CYCPAT_BM_READ);
            #endif
                // Make bank A0 a character pattern data read
                CYCLE_VDP_READ(CYCLE_A_REG);

            while (Switch_VBL_IN);

            #if 0
                // Handled in accordance with specific conditions
                slScrCycleSet(CYCPAT_BM_READ, CYCPAT_BM_READ, CYCPAT_CPU_RW,  CYCPAT_CPU_RW);
            #endif

            // Set bank B0 to CPU write mode
            CYCLE_CPU_WRITE(CYCLE_B_REG);

            for ( ; iu<movie_y; iu++)
            {
                DMA_ScuMemCopy(dst, src, movie_x_size);
                src += movie_x_size;
                dst += 2048;
            }

            #if 0
                // Defined for processing specific file formats
                slScrCycleSet(CYCPAT_BM_READ, CYCPAT_BM_READ, CYCPAT_BM_READ, CYCPAT_BM_READ);
            #endif
                // Make bank B0 a character pattern data read
                CYCLE_VDP_READ(CYCLE_B_REG);
            #endif

            // Special handling of the specific code
            CPK_CompleteDisp(cpk);
        }

        if (CPK_GetPlayStatus(cpk) == CPK_STAT_PLAY_END) // Movie End
        {
            // To handle the processing of the main function
            // Consideration of the main part of the data and its handling

            // Destroy Movie
            CPK_DestroyGfsMovie(cpk);

            GFS_Close(gfs);

            restart = 1;
        }
    }
}
